#!/usr/bin/env python
# -*- coding: utf-8; tab-width: 4; indent-tabs-mode: nil -*-
r'''a card file program with easy useful searching

Ramfile displays a text file as a collection of cards, which is useful
for finding bits of information after collecting and saving them.  To
find a card, type a word or words in the "Find" text box and press
Enter.  Ramfile finds the next card that contain all the words anywhere
on the card.  Unlike most other programs that search for strings exactly
as given, ramfile will find "john doe" even if the text is "doe, john".

Ramfile opens several file formats, including MacOS Ramfile / QuickFile,
mbox (mail), and any text file.  By default, case does not matter when
searching.  All open documents are searched.  The Find box highlightes
when a search has been all the way around.  Found words are highlighted
on the card.  Fairly large files can be handled; memory requirements are
about 1 to 1.4 times the size of the file, depending on the size of the
cards.

Command Line Usage:  ramfile [-hd] [<file>...]

  --geometry        X geometry option in pixels, e.g. 800x600.
  -h, --help        A brief usage message.
  -d, --no-default  Don't make default files "QuickData.txt" and
                    ".ramfile" (if they don't already exist).

With no file given, opens the default card file "~/QuickData.txt". 
Files are opened according to their apparent format: ramfile card file,
mbox (mail) file, or plain text.

Navigation:

The ramfile window is resizeable.  Under the menu bar is a popup to
indicate the way matching is done and a text entry field for the words
to search for ("the find box").  Under the find box is a notebook with a
tab for each document.  Each tab shows the document name and a close
box.  Tabs can be dragged to change their order or to move a document to
a new or different window.  Each tab's pane shows the scrollable,
editable text of the current card ("the card").  Move between the fields
with the mouse or the Tab key; alternate between the find box and the
card with Edit menu Find / Card (Control-F).

To move to the next (previous) card, press the Enter key (Shift+Enter)
with an empty find box.  To search forward (backward), type the words
you want to find and press Enter (Shift+Enter).  The matching words will
be highlighted on the card.  When a search has been all the way around,
the find box will highlight.  Searching is not case sensitive if the
words are all lower case.  Searching is actually done by string, not by
word, so matches may be inside words.  When more than one document is
open, all the documents are searched in sequence.

Click on a blue underligned link to find or open it.  Links starting
with a bullet ("•") are internal links that cause a search within the
open card documents.  Other links, either in angle brackets or otherwise
recognizable as being URLs, are opened as URLs, including email@address,
ftp.address, and www.web.address.

Return to the top of the document with Edit menu Go Home (not that the
top of a rolling file is all that important).  Add a new card after the
current card with Edit menu Add Card (Ctrl+R); delete the current card
with Edit menu Delete Card (Ctrl+L).

Details:

Ramfile understands a little bit about files from other OS's.  Files
with lines ending in "\r\n" are assumed to come from DOS / MSWindows and
are expected to be coded in latin1.  Files with lines ending in "\r" are
assumed to come from MacOS (or MOSX) and are expected to be coded in
mac-roman.  Otherwise, files are assumed to be in the current Locale's
encoding, often UTF-8.  More encoding stuff may follow.

In addition to ramfile's own card file format, ramfile can open mbox
(mail) files as cards.  Ramfile's ability to search for the search words
anywhere within individual messages is useful.  All the email headers
are a bit annoying; something may be done about that.  Ramfile can work
with reasonably large mbox files; my 1.2 GHz Athlon opens a 53 MB file
in about 7 seconds, and searches the document in about 2 seconds. 
Memory reqirements are about 1.4 times the file size plus a base (35 MB
on my system), some of that memory Python gives back when the document
is closed.

Several settings can be configured in ramfile's settings file
"~/.ramfile".  This is a file containing simple Python assignment
statements.  The settings are:

    dfltdoc     Pathnames for documents to open on startup, as a string
                for one pathname, or list of strings.
    geometry    A standard X Window Geometry specification, in pixels.
    url_opener  A pathname to the program to open URL links; if not
                available, Python's built-in stuff will be used.
    paste_odd   Menu items for the Edit menu Paste Odd Char submenu, as
                a list of tuples of strings: ( "name", "char(s)",
                "description").

Ramfile doesn't print yet.  Currently ramfile requires Python and GTK,
and opening MacOS files benefits from having my fastcharmap extension
module.  Printing would add other dependencies, as well as some
complexity in selecting what to print and how to lay out cards on a
page that I haven't yet had a need to figure out.

Ramfile doesn't have Undo.  GTK doesn't have Undo, and I'm not ready to
do that much work just now.  Ramfile does have File menu Revert and Edit
menu Revert Card.

Ramfile doesn't remember its window size and position, but you can use
the --geometry option to set them, say from a shortcut or Launcher icon,
or in the .ramfile settings file.
'''

__author__  = 'Tony Nelson'
__email__   = 'tonynelson@georgeanelson.com'
__url__     = 'http://georgeanelson.com/ramfile.htm'
__copyright__ = 'Copyright 2005, 2007 by George A. Nelson.  All rights reserved.'
__license__ = 'GPL'
__date__ = '24 Oct 2007'
__version__ = '0.1.3'
__credits__ = 'Bill Atkinson, author of QuickFile for MacOS.'


# nobug: TABing between find and card and back takes 3 TABs, as the
# notebook is one of the stops, and can be keyboard navigated with
# cursor keys.  Suck it up.
#   -- 4 TABs now that I've added the Find popup menu.

# fixed: the selection wasn't auto-scrolled into view.  Used the wrong
# call to add the TextView to the ScrolledWindow (add_with_viewport()
# rather than add()).  These all work now, but aren't needed:
#   sw = self.cur_page()
#   sw.page_view.move_mark_onscreen(sw.page_buffer.get_insert())
#   sw.page_view.place_cursor_onscreen()
#   sw.page_view.scroll_to_iter(sw.page_buffer.get_end_iter(), .3)
#   sw.page_view.scroll_to_mark(sw.page_buffer.get_insert(), .2)

# rfe: Should allow choosing the char encoding.  (I don't think
# auto-detecting the encoding is possible in most cases, as most
# encodings are just 8-bit codepages.  I note that gedit purports to
# autodetect the encoding but appears to only remember from last time.) 
# Right now, if the file line ending is CR, mac_roman encoding is
# assumed; if CR/LF, latin_1; otherwise UTF-8 is used.  Eventually there
# should be defaults for CR, CR/LF, and LF line endings.  There could be
# an encodings popup in the open and save dialogs, using the list at
# <http://docs.python.org/lib/standard-encodings.html> .

# partly done: have default encodings for Mac and Windows, in settings
# file, chosen according to the document's newlines.  What about the
# current locale, or whatever?

# rfe: MBOX importer (or something) should be able to show only
# "interesting headers", the way Eudora hides "boring headers".  There
# should be a way to, say, dial in and out of how boring one wants to
# see.  Subject: might be the least boring, then From:.  AFAIK, headers
# extend until the first blank line, and continue onto multiple lines by
# having leading whitespace (or TABs only?).  If we aren't allowing MBOX
# editing (we should mark the TextView or buffer as read-only), and
# don't allow doc.put() to do anything, we could just filter out lines
# on doc.get(), and when the dial setting changes.  Or use invisible
# style text.

# rfe: How about char stats:  chars, lines, cards?  Matched cards?

# rfe: Operations on the set of matched cards, such as make a new unsaved
# card document, or Save As Matched Only (option added to Save As dialog?).

# done: I18n.  Translatable strings are marked _(""" """).  Has been
# tried with Pig Latin (for the strings) and French (for the stock menu
# items).

# done: font / size choosing, sets current text field.

# rfe: redo the tabs when setting the font in the card.

# rfe: rebuild the sw.text_view tabs array when changing font.

# what about mapping diacritical marks, accents, etc, like case mapping?
#   -- not a Unicode concept.  Would have to invent something.  May be
#   possible to take makeunicodedata.py and ? and derive the chars that
#   decompose into 1 letter and some marks and do them.  Could also just
#   extend the Latin-1 sort of de-marking.  Not yet.  Result might be as
#   fast as just mapping to lower case, which it should also do.

# fixed: Accented characters don't even seem to do case-mapping.  Umm,
# str.lower() only does ASCII.  unicode.lower() does everything.

# rfe: Searching is 2x+ faster when not case mapping.  Could keep a copy
# of the mapped text; to save space it can be appended to the main text,
# keeping a note of the main text length.  In the case of mbox cards,
# which are about 4K avg, GNU malloc()'s rounding up to a page size
# accounts for the 1.4x memory usage.  Keeping 2 separate copies would
# mean 2.8x; appending would mean 2.4x.
#   -- 5 or 6 x slower now that I'm using unicode.lower().

# rfe: open directories:  ramfile and mbox open as separate docs; plain
# text opens as cards in a single doc, with the doc knowing the path to
# the dir and each card knowing its name and dirty state.  Save As could
# change between separate files and ramfile format.

# rfe: watch out for card marks inside cards?  Warn about split when
# card is saved?  Note that split won't happen until next time file is
# opened.  Some sort of escape mechanism?

# done: RE searching

# rfe: history popup of recent and saved search strings, along with
# their case mapping, accent mapping, and RE settings.  Make the "Find"
# label be a popup with check items for case mapping (lower() Find:
# find:), accent mapping (Fīnd: fīnd:), and RE (Re: Rē: re: rē:). 

# done: better URL links stuff.

# done: current help is pydoc format.  Should have a ramfile format help
# like Mac ramfile, now that hypertext links work.

# rfe: better help doc

# done: Show modified state in tab and window titles.

# rfe: Window menu?

# done: Show that no match was found.  Find area colored red on loop
# instead of green for found.

# done: More emphasis on whether Find area or Card is active.  Could be
# better.

# bug? If the search word is only found on the current card and it is
# selected, it won't be drawn hilighted, though it will be unselected.

# fixed: Revert doesn't set card correctly.

# fixed: Delete Card doesn't set card correctly.

# fixed: Add Card should focus on card.

# fixed: when searching, the old current card value is searched, not its
# edited value.  Probably should always save card before searching? 
# Think about it.

# fixed: should scroll to top of card when changing cards.

#rfe: parse dates and have a (use the?) calendar.  Gnome puts a calendar
#   under the menu bar clock, which is used by Evolution.  Don't know if
#   I can use it myself.
#   
#   import locale, time -- actually, should be a new easydate module
#   with a strptime() that takes whatever the current locale dates (and
#   times) can look like, allowing abbreviations, order deviations, and
#   omissions.  Omitted values will default to None, not some useless
#   wrong number, so they can easily be found and replaced with the
#   appropriate (current) value.  Order deviations should accept 1..31
#   as day-of-month and other numbers as year, or 00..99 as year in
#   common century (69..99 -> 1969..1999, 00..68 -> 2000..2068).  Day
#   and month abbreviations should be either the locale abbreviations,
#   or, longer than the abbreviations and unambiguous.  Return a
#   standard 9-tuple (y,m,d,h,m,s,wd,yd,dst), with unspecified values
#   set to None (except DST, which will be -1 for ease of use).  I can
#   test in en_US.UTF-8, en_CA.UTF-8, en_GB.UTF-8, and fr_FR.UTF-8. 
#   French will have UTF-8 abbreviations longer than 3 bytes.

# fixed: new files should have platform's newlines

# fixed: sys.path.insert() takes 2 arguments


import os
import sys
import new
import re
import time
import stat
import platform


us = os.path.splitext(os.path.basename(sys.argv[0]))[0]
#usdir = os.path.dirname(sys.argv[0])
usdir = os.path.abspath(os.path.dirname(__file__))
home = '~'
if platform.system().lower() == 'windows':
    winhome = os.path.join(home, 'My Documents')
    if os.path.isdir(os.path.abspath(os.path.expanduser(winhome))):
        home = winhome

#def _(str): return str
import gettext
transdir = os.path.join(usdir, 'locale')
gettext.install(us, transdir)
#lang1 = gettext.translation( 'ramfile', localedir=transdir,
#                             languages=['en_PL'] )
#lang1.install()


def get_default_settings():
    return _(
        """# Ramfile settings are Python statements, executed with some protection.\n"""
        """\n"""
        """# The default settings are:\n"""
        """#\n"""
        """dfltdoc = '~/QuickData'\n"""
        """geometry = '400x380'                # in pixels\n"""
        """url_opener = '/usr/bin/gnome-open'  # fallback is webbrowser module\n"""
        """pasteOdd = [                        # [(menu item, odd text, tooltip),]\n"""
        """    ('∂','∂','Date mark'),\n"""
        """    ('•','•','Internal Link mark'),\n"""
        """    ('_Tab','\\t','Tab character'),\n"""
        """    ('_NBSP','\\xC2\\xA0','Non-breaking space'),\n"""
        """  ]\n"""
        """font = ''                           # system default font\n"""
        """mac_encoding = 'mac_roman'          # when doc's newlines are RETURN\n"""
        """win_encoding = 'latin_1'            # when doc's newlines are RETURN LINEFEED\n"""
        """\n"""
        """# Some sample statements:\n"""
        """#\n"""
        """#dfltdoc = ['ramfile.py', '~/QuickData'] # dfltdoc can be a list\n"""
        """#print "hello from .ramfile" # many statements are legal, if not useful\n"""
        """#pasteOdd = ['a', ('_Percent','%','A sample tip'), 'hello']\n"""
        """#font = 'Serif Italic 10'\n""" )

dfltdoc = orig_dfltdoc = os.path.join(home, 'QuickData.txt')
geometry = _("""480x380""")
url_opener = '/usr/bin/gnome-open'
pasteOdd = [                # (character, name+char, text)
    ('∂','∂','Date mark'),
    ('•','•','Internal Link mark'),
    ('_Tab','\t','Tab character'),
    ('_NBSP','\xC2\xA0','Non-breaking space'),
  ]
font = _("""""")            # use system default font
mac_encoding = _("""mac_roman""")
win_encoding = _("""latin_1""")

opt_no_default_files = False

help_file = _("""ramfile-help""")
help_geometry = _("""480x380-0""")
help_tab_chars = int(_("""24"""))


class Go:
    down = 1 ; up = -1

class Phase:
    all, next, rest = range(3)

class Matching:
    names = _("""FIND:;find:;Find:;RE:;re:;Re:""").split(';')
    exact, casemap, auto, _n = range(4)
    
    def __init__(self, mapping=auto, use_regex=False):
        self.mapping = mapping
        self.use_regex = use_regex
    
    def item_n(self):
        return self.mapping + self.use_regex*self._n
    
    def set(self, n):
        self.mapping = n % 3
        self.use_regex = n >= 3

class Error(Exception):
    def __init__(self, message):    self.message = message
    def __str__(self):              return self.message

def my_tempnam(path):
    '''Make a temp file path.
    
    os.tempnam() can't be used as it might not use the passed dirPath,
    and we should make our files in the same directory as the document
    destination.
    '''
    for i in xrange(10000):
        suffix = '.tmp%04d' % i
        if not os.path.exists(path + suffix):
            return path + suffix
    raise Error('''my_tempnam(%s) can't make new name''' % path)

def get_settings():
    sys.modules['settings'] = new.module('settings')
    import settings
    path = os.path.expanduser('~/.'+us)
    try:
        f = open(path)
        s = f.read()
        f.close()
    except IOError:
        s = get_default_settings()
        path = _("""the built-in settings""")
    try:
        exec s in settings.__dict__
    except:
        print _("""%s: trouble with settings file "%s": %s""") % (
            us, path, str(sys.exc_info()[1]) )
    else:
        global dfltdoc, geometry, url_opener, pasteOdd, font
        global mac_encoding, win_encoding
        def get_setting(name, types):
            temp = getattr(settings, name, globals()[name])
            for t in types + [type(None)]:
                if isinstance(temp, t):
                    globals()[name] = temp
                    break
        get_setting('dfltdoc',      [str, list])
        get_setting('geometry',     [str,     ])
        get_setting('url_opener',   [str,     ])
        get_setting('pasteOdd',     [     list])
        get_setting('font',         [str,     ])
        get_setting('mac_encoding', [str,     ])
        get_setting('win_encoding', [str,     ])

def make_default_settings():
    path = os.path.expanduser('~/.'+us)
    if not os.path.exists(path):
        f = open(path, 'w')
        f.write(get_default_settings())
        f.close()

def my_import(name):
    try:
        mod = __import__(name)
    except ImportError:
        return None
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


class CardDoc:
    '''Card collection document.
    
    Documents are not aware of GTK.  They implement loading and saving
    through Importers, handle encoding conversions during load and save,
    and get, put, revert, and find cards. 
    '''
    ld_ok, ld_fail, ld_no_file, ld_enc_err = range(4)
    
    def opendoc(path):
        '''Factory:  make doc by pathname.
        '''
        try:
            f = open(path,'rU',256*1024)
        except IOError:
            return None, CardDoc.ld_no_file
        doc = None
        ld = CardDoc.ld_fail
        for importer in importers:
            if importer.hasFormat(f):
                doc = CardDoc() ### could be "importer.make_doc(), which could make a subclass of CardDoc
                if doc:
                    ld = doc.load(f, path, importer)
                break
        f.close()
        return doc, ld
    opendoc = staticmethod(opendoc)

    def newdoc():
        '''Factory:  make empty Quickfile doc.
        '''
        doc = CardDoc()
        if doc:
            doc.cards.append('')
            doc.importer = QuickfileImporter;
            doc.newlines = os.linesep
        return doc
    newdoc = staticmethod(newdoc)
    
    def __init__(self):
        self.cards = []
        self.path = ''
        self.mode = 0644
        self.importer = None
        self.encoding = 'utf-8'
        self.newlines = None
        self.curCard = 0
        self.loopCard = 0
        self.dirty = False
        self.readonly = False
    
    def load(self, f, path, importer):
        self.cards = []
        self.path = path
        self.importer = importer
        self.mode = os.stat(path).st_mode
        self.encoding = 'utf-8'
        self.curCard = 0
        self.loopCard = 0
        self.dirty = False
        self.readonly = not os.access(path, os.W_OK)
        if isinstance(f.newlines, str):
            self.newlines = f.newlines
        no_fastcharmap = False
        st = time.time()
        importer.read(self, f)
        rt = time.time()
        if isinstance(self.newlines, str):
            if self.newlines == '\r':      # Mac newline
                self.encoding = mac_encoding
                # Fix missing Apple logo in mac_roman (and other Mac encodings?).
                #import encodings.mac_roman
                enc = my_import('encodings.%s' % self.encoding)
                if hasattr(enc, 'decoding_map'):
                    if not enc.decoding_map[0xF0]:
                        enc.decoding_map[0xF0] = 0xF8FF
                        enc.encoding_map[0xF8FF] = 0xF0
                    try:
                        import fastcharmap
                    except ImportError:
                        no_fastcharmap = True
                    else:
                        fastcharmap.hook(self.encoding)
            if self.newlines == '\r\n':    # DOS newline
                self.encoding = win_encoding
                # I think latin_1 is already optimized -- latin_n may not be
                enc = my_import('encodings.%s' % self.encoding)
                if hasattr(enc, 'decoding_map'):
                    try:
                        import fastcharmap
                    except ImportError:
                        no_fastcharmap = True
                    else:
                        fastcharmap.hook(self.encoding)
            #for i in xrange(len(self.cards)):
            #    u = unicode(self.cards[i], encoding)
            #    self.cards[i] = u.encode('utf_8')
            try:                        # decode to or validate as UTF-8
                for i in xrange(len(self.cards)):
                    u = unicode(self.cards[i], self.encoding)
                    if self.encoding != 'utf-8':
                        self.cards[i] = u.encode('utf_8')
            except UnicodeDecodeError:
                for i in xrange(len(self.cards)):
                    u = unicode(self.cards[i], self.encoding, 'replace')
                    self.cards[i] = u.encode('utf-8')
                return CardDoc.ld_enc_err
        ct = time.time()
        if ct-st > .2:
            print '%s: cards %d  read time %6.2f  convert time %6.2f' % (
                us, len(self.cards), rt-st, ct-rt )
            if no_fastcharmap:
                print _("""    charmap codec used but fastcharmap not available""")
        if not len(self.cards):
            self.cards.append('')
        return CardDoc.ld_ok    # note alternate return in "except UnicodeDecodeError:"
    
    def save(self):
        '''Save doc to a file.
        
        All saves are save-as, either to a new file or a temp file, for
        safety. 
        '''
        if self.readonly or not self.dirty:
            return
        tmpPath = my_tempnam(self.path)
        f = open(tmpPath,'wb',256*1024)
        os.chmod(tmpPath, 0600)
        st = time.time()
        self.importer.write(self, f)
        f.close()
        wt = time.time()
        backup = self.path+'~'
        if os.path.isfile(backup):
            os.remove(backup)
        if os.path.isfile(self.path) and not os.path.exists(backup):
            os.rename(self.path, backup)
        if not os.path.exists(self.path):
            os.rename(tmpPath, self.path)
            os.chmod(self.path, self.mode)
        self.dirty = False
        if wt-st > .2:
            print '%s: cards %d  write time %6.2f' % (
                us, len(self.cards), wt-st)
    
    def save_as(self, path):
        print path
        self.readonly = False       # ?? not os.access(path, os.W_OK)
        self.path = path
        self.dirty = True
        self.save()
    
    def revert(self):
        if not self.path or not self.dirty:
            return
        f = open(self.path,'rU',256*1024)
        if self.load(f, self.path, self.importer) != CardDoc.ld_ok:
            pass # Umm, something bad happened.
        f.close()
    
    def find(self, findStr, direction, matching, phase):
        '''Find words in findstr -> (cardNum, looped).
        '''
        # parse search words from str
        #   if empty, go one card in direction
        # if no caps, case insensitive search
        # phases:   Go.all      from "start" to   "end"
        #           Go.next     from "next"  to   "end"
        #           Go.rest     from "start" thru "orig"
        
        
        #   Search all open card docs as if laid end-to-end in a continuous
        #   scroll.  In the current direction, start from the "next" card in the
        #   current doc, go to the "end", then go to the "next" file, start from
        #   the "top", do all of it, then each other file the same, then the
        #   first file, from the "top", go to the start card.  Looping is when,
        #   during several of these, the search passes the original start card. 
        #   Looping does not affect the search; the user is merely notified in
        #   some way.  Searching now has 3 phases:
        #
        #       from "next" to "end"
        #       from "start" to "orig"
        #       from "start" to "end"
        #
        #   The direction can be up or down.  There is still a current card for
        #   each document:  the last card found there (but not necessarily from
        #   the current search).  Switching documents should show a sensible
        #   card.  Searches should always land on a sensible card:  either the
        #   found card or the start card.  In an unsuccessful search, the
        #   current card is not changed; note that for an unsucessful "start" to
        #   "end" search, the current card is also not changed even though it
        #   wasn't the start of the search in that doc.
        
        start = 0
        end = len(self.cards)
        if direction == Go.up:
            start, end = end-1, -1
        ci = start
        if phase == Phase.next:
            ci = self.curCard + direction
        elif phase == Phase.rest:
            end = self.curCard + direction
        loopCard = self.loopCard
        re_ = None
        if findStr and matching.use_regex:
            try:
                re_ = re.compile(findStr)
            except re.error:
                return (None, False)
        else:    
            words = findStr.split()
            if not len(words):
                looped = (ci == loopCard)
                if ci == end:
                    ci = None
                return (ci, looped)
        if matching.mapping == Matching.auto:
            ufindStr = unicode(findStr, 'utf-8')
            noCase = ufindStr.lower() == ufindStr
        elif matching.mapping == Matching.casemap:
            noCase = True
        else: # matching.mapping == Matching.exact:
            noCase = False
        if noCase and not re_:  # convert to Unicode once <1% faster
            words = [unicode(word, 'utf-8') for word in words]
        looped = False
        while ci != end:
            if ci == loopCard:
                looped = True
            card = self.cards[ci]
            if noCase:  # vs s.lower() unicode() 10% slower; u.lower() >3x slower
                card = unicode(card, 'utf-8').lower() # and >6x slower than exact matching
            if re_:
                found = (re_.search(card) != None)
            else:
                found = True
                for word in words:
                    if card.find(word) < 0:
                        found = False
                        break;
            if found:
                return ci, looped
            ci += direction
        return None, looped
    
    def find_positions(self, card, findStr, matching):
        '''Find words as in find() -> [ (pos, len)... ]
        '''
        re_ = None
        if findStr and matching.use_regex:
            try:
                re_ = re.compile(findStr)
            except re.error:
                return []
        else:    
            words = findStr.split()
            if not len(words):
                return []
        if matching.mapping == Matching.auto:
            noCase = findStr.lower() == findStr
        elif matching.mapping == Matching.casemap:
            noCase = True
        else: # matching.mapping == Matching.exact:
            noCase = False
        if noCase:
            card = unicode(card, 'utf-8').lower().encode('utf-8')
        wl = []
        if re_:
            for m in re_.finditer(card):
                wl.append((m.start(), m.end()-m.start()))
        else:
            for word in words:
                p = 0
                l = len(word)
                while True:
                    p = card.find(word, p)
                    if p < 0:
                        break;
                    wl.append((p, l))
                    p += l
        return wl
    
    def get(self, cardNum):
        self.range_check(cardNum, 'get')
        return self.cards[cardNum]
    
    def get_save(self, cardNum):
        self.range_check(cardNum, 'get')
        card = self.cards[cardNum]
        if self.encoding and self.encoding != 'utf-8':
            u = unicode(card, 'utf-8')
            card = u.encode(self.encoding)
        return card
    
    def put(self, card, cardNum):
        self.range_check(cardNum, 'put')
        if self.readonly:
            return
        if not len(card) or card[-1] != '\n':
            card += '\n'
        self.cards[cardNum] = card
        self.dirty = True
    
    def add(self, cardNum):
        self.range_check(cardNum, 'add', 0)
        if self.readonly:
            return
        self.cards.insert(cardNum, '')
    
    def delete(self, cardNum):
        self.range_check(cardNum, 'delete')
        if self.readonly:
            return
        del self.cards[cardNum]
        if not len(self.cards):
            self.cards.append('')
        if self.curCard >= len(self.cards):
            self.curCard = 0    # wrap around: always show the "next card"
    
    def range_check(self, cardNum, who, fence=-1):
        if 0 > cardNum > len(self.cards)+fence:
            raise Error(  _("""CardDoc.%s(%d) out of range 0..%d""")
                        % (who, cardNum, len(self.cards)+fence) )
    
    def set_curCard(self, cardNum):
        if cardNum < 0:
            cardNum = 0
        if cardNum > len(self.cards)-1:
            cardNum = len(self.cards)-1
        self.curCard = cardNum


class Importer:
    '''Importer base class
    
    Card file importers / exporters recognize, read, or write card
    files.  If I'm doing this right, Importers can also be modules with
    these functions.
    
    Importers re-read part of the file, so many flo's won't work.
    '''
    
    def hasFormat(f):
        '''True if f is this format.
        
        Seek to 0 first.  Please read 3 lines in order to prime f.newlines
        attribute.
        '''
        return False
    hasFormat = staticmethod(hasFormat)
    
    def read(doc, f):
        raise Error('Importer.read() unimplemented')
    read = staticmethod(read)
    
    def write(doc, f):
        raise Error('Importer.write() unimplemented')
    write = staticmethod(write)

class QuickfileImporter(Importer):
    r'''Quickfile format
    
    Cards are separated by a line == "*".  The file starts and ends with
    "\n*\n*\n" (two blank cards), which is removed on import and
    replaced on export.
    '''
    
    def hasFormat(f):
        try:
            f.seek(0)
            if ( f.readline() == '\n'
             and f.readline() == '*\n'
             and f.readline() == '*\n'):
                return True
        except IOError:
            pass
        return False
    hasFormat = staticmethod(hasFormat)
    
    def read(doc, f):
        f.seek(0) ; f.readline() ; f.readline() ; f.readline()
        cardlines = []
        for l in f:
            if l == '*\n':
                card = ''.join(cardlines)
                doc.cards.append(card)
                cardlines = []
            else:
                cardlines.append(l)
        if len(cardlines):
            card = ''.join(cardlines)
            doc.cards.append(card)
        if doc.cards[-1] == '':  del doc.cards[-1]
        if doc.cards[-1] == '':  del doc.cards[-1]
    read = staticmethod(read)
    
    def write(doc, f):
        nl = '\n'
        if isinstance(doc.newlines, str):
            nl = doc.newlines
        f.write(nl+'*'+nl+'*'+nl)
        for i in xrange(len(doc.cards)):
            card = doc.get_save(i)
            if nl != '\n':
                card = card.replace('\n', nl)
            f.write(card)
            if not card.endswith(nl):
                f.write(nl)
            f.write('*'+nl)
        f.write('*'+nl)
    write = staticmethod(write)

class MboxImporter(Importer):
    '''Mbox format
    
    Cards are separated by a line starting "From ".
    '''
    key = 'From '
    
    def hasFormat(f):
        try:
            f.seek(0)
            l = f.readline()
            l1 = f.readline()
            if not l:  l = l1
            l1 = f.readline()
            if not l:  l = l1
            if l.startswith(MboxImporter.key):
                return True
        except IOError:
            pass
        return False
    hasFormat = staticmethod(hasFormat)
    
    def read(doc, f):
        f.seek(0)
        key = MboxImporter.key
        keylen = len(key)
        cardlines = []
        for l in f:
            if ( l.startswith(key) and len(cardlines) ):
                card = ''.join(cardlines)
                doc.cards.append(card)
                cardlines = []
            cardlines.append(l)
        if len(cardlines):
            card = ''.join(cardlines)
            doc.cards.append(card)
    read = staticmethod(read)
    
    def write(doc, f):
        nl = '\n'
        if isinstance(doc.newlines, str):
            nl = doc.newlines
        for i in xrange(len(doc.cards)):
            card = doc.get_save(i)
            if nl != '\n':
                card = card.replace('\n', nl)
            f.write(card)
            if not card.endswith(nl):
                f.write(nl)
    write = staticmethod(write)
    # also used by FileImporter!

class FileImporter(Importer):
    '''Default format
    
    The entire file is a single card.
    '''
    
    def hasFormat(f):
        try:
            f.seek(0)
            f.readline()
            f.readline()
            f.readline()
            return True
        except IOError:
            pass
        return False
    hasFormat = staticmethod(hasFormat)
    
    def read(doc, f):
        f.seek(0)
        cardlines = []
        for l in f:
            cardlines.append(l)
        if len(cardlines):
            card = ''.join(cardlines)
            doc.cards.append(card)
    read = staticmethod(read)
    
    write = staticmethod(MboxImporter.write)

importers = [QuickfileImporter, MboxImporter, FileImporter]


def cli():
    '''Command-Line Interface
    
    Return list of document paths or None, or exit on bad parameters.
    '''
    import getopt
    def get_usage():
        return _(
            """%s %s  usage: %s [-hd] [<file>...]\n"""
            """A card file program with easy useful searching.\n"""
            """\n"""
            """  --geometry        X geometry option in pixels, e.g. 800x600.\n"""
            """  -h, --help        This message.\n"""
            """  -d, --no-default  Don't make default files "%s" and\n"""
            """                    ".%s" (if they don't already exist).\n"""
            """\n"""
            """With no file opens the default card file "%s".  Files are opened\n"""
            """according to their apparent format: %s card file, mbox (mail) file,\n"""
            """or plain text.\n""" ) % (
            us, __version__, us, dfltdoc, us, dfltdoc, us )
    global geometry, opt_no_default_files
    
    # Scripts don't get the current directory in their path by default.
    # (Actually, what we might need is /this/ script's directory.)
    scriptdir = os.path.dirname(sys.argv[0])
    if scriptdir in sys.path:
        sys.path.remove(scriptdir)
    #sys.path.insert(0, '.')
    sys.path.insert(0, usdir)
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'hd', ['geometry=', 'help', 'no-default'])
        for opt, val in opts:
            if opt == '--geometry':
                geometry = val
            if opt in ['--help', '-h']:
                print get_usage()
                sys.exit(0)
            if opt in ['--no-default', '-d']:
                opt_no_default_files = True
        
        #should find other importers in various places
        
        files = []
        for arg in args:
            if not os.path.isfile(arg):
                print _("""%s: no file "%s" """) % (us, arg)
            else:
                files.append(arg)
        if len(files):
            return files
    except getopt.error, (msg, opt):
        print '%s: %s' % (us, msg)
        print _("""Try "%s -h" or "pydoc %s" for more information.""") % (
            us, os.path.splitext(us)[0] )
        sys.exit(1)

def main():
    get_settings()
    files = cli()
    ok_no_file = False
    global dfltdoc
    if not files and dfltdoc:
        if isinstance(dfltdoc, str):
            dfltdoc = [ dfltdoc ]
        files = []
        for path in dfltdoc:
            path = os.path.expanduser(path)
            path = os.path.abspath(path)
            if os.path.isfile(path):
                files.append(path)
    if not files:
        if opt_no_default_files:
            ok_no_file = True
        else:
            path = None
            if dfltdoc:
                path = dfltdoc[0]
            if not path:
                path = orig_dfltdoc
            path = os.path.expanduser(path)
            path = os.path.abspath(path)
            if not os.path.exists(path):
                path1 = os.path.splitext(orig_dfltdoc)[0]
                path1 = os.path.expanduser(path1)
                path1 = os.path.abspath(path1)
                if os.path.exists(path1):
                    path = path1
            if os.path.exists(path):
                files = [path]
            else:
                doc = CardDoc.newdoc()
                doc.path = path
                doc.dirty = True
                try:
                    doc.save()
                except:
                    print _("""%s: can't create "%s" """) % (us, path)
                else:
                    files = [path]
    import ramfile_ui_gtk as gui
    w = gui.RamfileWind()
    w.set_sensitive(False)
    gui.gtkeat()
    opened = False
    nextTime = time.time() + .2
    for path in files:
        path = os.path.abspath(path)
        doc, ld = CardDoc.opendoc(path)
        if ld == CardDoc.ld_enc_err:
            print _("""%s: Encoding error: file "%s" is not %s.""") % (
                us, path, doc.encoding )
        elif ld == CardDoc.ld_no_file:
            print _("""%s: can't open "%s" """) % (us, path)
        elif doc:
            w.make_doc_view(doc)
            opened = True
        curTime = time.time()
        if curTime > nextTime:
            gui.gtkeat()
            nextTime = curTime + .2
    if not opened and not ok_no_file:
        sys.exit(1)
    if not opt_no_default_files:
        make_default_settings()
    w.set_sensitive(True)
    w.find_edit.grab_focus()
    gui.main()

if __name__ == '__main__':
    main()
